;;; sdiff --- S-expression diff
;;; Copyright © 2021 Arun Isaac <arunisaac@systemreboot.net>
;;;
;;; This file is part of sdiff.
;;;
;;; sdiff is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; sdiff is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with sdiff.  If not, see <https://www.gnu.org/licenses/>.

(define-module (hungarian)
  #:use-module ((rnrs base) #:select (infinite?))
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-43)
  #:use-module (ice-9 match)
  #:use-module (matrix)
  #:export (hungarian
            minimum-cost-edge-cover))

(define (exact-minimum x y)
  "Return the minimum of X and Y. X, Y and the return value are exact
numbers or an infinite value."
  (let ((result (min x y)))
    (if (infinite? result)
        result
        (inexact->exact result))))

;; This implementation of the Hungarian algorithm follows the
;; description at https://brc2.com/the-algorithm-workshop/
(define (hungarian adjacency-input)
  (define k
    (match (array-dimensions adjacency-input)
      ((rows cols) (min rows cols))))

  ;; Rotate the adjacency matrix such that there are at least as many
  ;; columns as there are rows.
  (define adjacency
    (match (array-dimensions adjacency-input)
      ((rows cols)
       (let ((matrix (make-array 0 (min rows cols) (max rows cols))))
         (array-index-map! matrix (lambda (i j)
                                    (if (<= rows cols)
                                        (array-ref adjacency-input i j)
                                        (array-ref adjacency-input j i))))
         matrix))))
  
  (define mask (apply make-array #f (array-dimensions adjacency)))
  (define row-cover (make-array #f (match (array-dimensions adjacency)
                                     ((rows cols) rows))))
  (define column-cover (make-array #f (match (array-dimensions adjacency)
                                        ((rows cols) cols))))

  (define (find-zero-with-mask sub-adjacency mask-state)
    (matrix-any (lambda (i j element)
                  (and (zero? element)
                       (eq? (array-ref mask i j) mask-state)
                       (cons i j)))
                sub-adjacency))

  (define (find-starred-zero sub-adjacency)
    (find-zero-with-mask sub-adjacency 'star))

  (define (find-primed-zero sub-adjacency)
    (find-zero-with-mask sub-adjacency 'prime))

  (define (covered? i j)
    (or (array-ref row-cover i)
        (array-ref column-cover j)))

  (define (find-uncovered-zero)
    (matrix-any (lambda (i j element)
                  (and (zero? element)
                       (not (covered? i j))
                       (cons i j)))
                adjacency))

  (define (number-of-covered-columns)
    (vector-fold (lambda (_ result element)
                   (if element (1+ result) result))
                 0
                 column-cover))

  (define (step-1)
    ;; Subtract the smallest element of each row from that row.
    (for-matrix-row (lambda (_ row)
                      (array-map! row (cute - <> (matrix-min row)) row))
                    adjacency)
    (step-2))

  (define (step-2)
    ;; Star every zero that does not have another starred zero in its
    ;; row or column.
    (array-index-map! mask (lambda (i j)
                             (if (and (zero? (array-ref adjacency i j))
                                      (not (find-starred-zero (matrix-row adjacency i)))
                                      (not (find-starred-zero (matrix-col adjacency j))))
                                 'star
                                 (array-ref mask i j))))
    (step-3))

  (define (step-3)
    ;; Cover each column containing a starred zero.
    (for-matrix-col (lambda (column-index column)
                      (when (find-starred-zero column)
                        (array-set! column-cover #t column-index)))
                    adjacency)
    ;; We are done if all columns are covered.
    (unless (= k (number-of-covered-columns))
      (step-4)))

  (define (step-4)
    ;; Find an uncovered zero.
    (match (find-uncovered-zero)
      ((zero-row . zero-column)
       ;; Prime the found uncovered zero.
       (array-set! mask 'prime zero-row zero-column)
       ;; Find a starred zero in the row containing the primed zero.
       (match (find-starred-zero (matrix-row adjacency zero-row))
         ;; Cover this row and the column containing the starred zero.
         ((starred-zero-row . starred-zero-column)
          (array-set! row-cover #t zero-row)
          (array-set! column-cover #f starred-zero-column)
          (step-4))
         ;; If there is no starred zero, go to step 5.
         (#f (step-5 (cons zero-row zero-column)))))
      ;; If there is no uncovered zero, go to step 6.
      (#f (step-6))))

  (define (step-5 start)
    ;; Find a series of alternating primed and starred zeros. Unstar
    ;; starred zeros, and star primed zeros.
    (for-each (match-lambda
                ((i . j)
                 (array-set! mask
                             (case (array-ref mask i j)
                               ((star) #f)
                               ((prime) 'star))
                             i j)))
              (step-5-find-series start))
    ;; Clear all covers.
    (array-map! row-cover (const #f))
    (array-map! column-cover (const #f))
    ;; Erase all primes.
    (array-map! mask (lambda (element)
                       (if (eq? element 'prime)
                           #f element))
                mask)
    (step-3))

  (define (step-5-find-series start)
    (cons start
          (match start
            ((zero-row . zero-column)
             (case (array-ref mask zero-row zero-column)
               ((prime)
                (cond
                 ((find-starred-zero (matrix-col adjacency zero-column))
                  => step-5-find-series)
                 (else (list))))
               ((star)
                (step-5-find-series
                 (find-primed-zero (matrix-row adjacency zero-row)))))))))

  (define (step-6)
    (let ((smallest-uncovered (matrix-fold (lambda (i j element result)
                                             (if (covered? i j)
                                                 result
                                                 (exact-minimum element result)))
                                           +inf.0
                                           adjacency)))
      ;; Add the smallest uncovered element to every element of each
      ;; covered row.
      (for-matrix-row (lambda (row-index row)
                        (when (array-ref row-cover row-index)
                          (array-map! row (cut + <> smallest-uncovered) row)))
                      adjacency)
      ;; Subtract the smallest uncovered element from every element of
      ;; each uncovered column.
      (for-matrix-col (lambda (column-index column)
                        (unless (array-ref column-cover column-index)
                          (array-map! column (cut - <> smallest-uncovered) column)))
                      adjacency)
      (step-4)))

  (step-1)
  (map (match-lambda
         ((i . j)
          ;; Transpose assignments if original adjacency matrix was
          ;; transposed.
          (match (array-dimensions adjacency-input)
            ((rows cols)
             (if (<= rows cols)
                 (cons i j)
                 (cons j i))))))
       ;; Pick out starred zeros for assignment.
       (matrix-fold (lambda (i j element result)
                      (if (and (zero? element)
                               (array-ref column-cover j)
                               (eq? (array-ref mask i j) 'star))
                          (cons (cons i j) result)
                          result))
                    (list)
                    adjacency)))

;; This algorithm for finding the minimum cost edge cover is described
;; in "§19.3 Finding a minimum-weight edge cover" on page 317 of the
;; [2004] book "Combinatorial Optimization: Polyhedra and Efficiency
;; (Volume A)" by Alexander Schrijver.
(define (minimum-cost-edge-cover adjacency)
  "Return the minimum cost edge cover of the graph represented by
ADJACENCY matrix."
  (match (array-dimensions adjacency)
    ((rows cols)
     (let ((larger-adjacency (make-array 0 (+ rows cols) (+ rows cols))))
       (array-index-map! larger-adjacency
                         (lambda (i j)
                           (cond
                            ;; Original graph
                            ((and (< i rows) (< j cols))
                             (array-ref adjacency i j))
                            ;; Edges between original graph and its copy
                            ((and (>= i rows) (< j cols))
                             (if (= (- i rows) j)
                                 (* 2 (matrix-min (matrix-col adjacency j)))
                                 +inf.0))
                            ((and (< i rows) (>= j cols))
                             (if (= i (- j cols))
                                 (* 2 (matrix-min (matrix-row adjacency i)))
                                 +inf.0))
                            ;; Copy of original graph
                            ((and (>= i rows) (>= j cols))
                             (array-ref adjacency (- j cols) (- i rows))))))
       (delete-duplicates
        (filter-map (match-lambda
                      ((i . j)
                       (cond
                        ;; Keep edges in the original graph.
                        ((and (< i rows) (< j cols))
                         (cons i j))
                        ;; Replace edges between original graph and its copy.
                        ((and (>= i rows) (< j cols))
                         (matrix-fold (lambda (i j element result)
                                        (match result
                                          ((result-i . result-j)
                                           (if (< element
                                                  (array-ref adjacency result-i result-j))
                                               (cons i j)
                                               result))))
                                      (cons 0 j)
                                      (matrix-col adjacency j)))
                        ((and (< i rows) (>= j cols))
                         (matrix-fold (lambda (i j element result)
                                        (match result
                                          ((result-i . result-j)
                                           (if (< element
                                                  (array-ref adjacency result-i result-j))
                                               (cons i j)
                                               result))))
                                      (cons i 0)
                                      (matrix-row adjacency i)))
                        ;; Delete edges in the copy of the original graph.
                        ((and (>= i rows) (>= j cols)) #f))))
                    (hungarian larger-adjacency)))))))
