;;; sdiff --- S-expression diff
;;; Copyright © 2021 Arun Isaac <arunisaac@systemreboot.net>
;;;
;;; This file is part of sdiff.
;;;
;;; sdiff is free software: you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; sdiff is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with sdiff.  If not, see <https://www.gnu.org/licenses/>.

(define-module (sdiff)
  #:use-module ((rnrs base) #:select (finite? let-values))
  #:use-module (srfi srfi-1)
  #:use-module (srfi srfi-26)
  #:use-module (srfi srfi-171)
  #:use-module (ice-9 arrays)
  #:use-module (ice-9 match)
  #:use-module (ice-9 pretty-print)
  #:use-module (oop goops)
  #:use-module (hungarian)
  #:use-module (matrix)
  #:export (sdiff))

;; This file implements the MH-DIFF algorithm described in "Meaningful
;; Change Detection in Structured Data" by Sudarshan S. Chawathe and
;; Hector Garcia-Molina. The paper is available at
;; http://ilpubs.stanford.edu:8090/140/1/1996-15.pdf

;; Deleting a node and inserting its replacement is more expensive
;; than just updating it.
(define cost-insert 2)
(define cost-delete cost-insert)

(define (cost-update m n tree1 tree2)
  (if (equal? (tree-label tree1 m)
              (tree-label tree2 n))
      0
      3))

;; Moving, coping and gluing are essentially prohibited by setting
;; their cost very high.
(define cost-move 100)
(define cost-copy 100)
(define cost-glue cost-copy)

(define-class <sdiff-node> ()
  (tag #:init-keyword #:tag #:getter sdiff-node-tag)
  (label #:init-keyword #:label #:getter sdiff-node-label))

(define-method (write (node <sdiff-node>) port)
  (unless (eq? (sdiff-node-tag node) 'fake)
    (let ((red 31)
          (green 32))
      (format port "~a[~am" #\esc (case (sdiff-node-tag node)
                                    ((insert update-new) green)
                                    ((delete update-old) red)))
      (write (sdiff-node-label node) port)
      (format port "~a[0m" #\esc))))

;; TODO: Any two fake nodes are equal. This should not be.
(define-method (equal? (node1 <sdiff-node>) (node2 <sdiff-node>))
  (and (eq? (sdiff-node-tag node1)
            (sdiff-node-tag node2))
       (equal? (sdiff-node-label node1)
               (sdiff-node-label node2))))

(define (make-proper-tree tree)
  (match tree
    (((? list? parent) children ...)
     (make-proper-tree (cons (make <sdiff-node> #:tag 'fake #:label 'fake)
                             tree)))
    ((parent children ...)
     (cons parent (map make-proper-tree children)))
    (leaf leaf)))

(define (nodeset tree)
  "Return list of nodes in proper TREE. The returned list of nodes are
arranged in pre-order traversal order."
  (match tree
    ((parent children ...)
     (cons parent (append-map nodeset children)))
    (leaf (list leaf))))

(define (pre-order-traversal tree)
  "Return TREE where every node is replaced by its pre-order traversal
index. The pre-order traversal index is 0 for the root node."
  (define (pre-order-traversal-recursive tree index)
    (match tree
      ((parent children ...)
       (match (fold (match-lambda*
                      ((child (index result))
                       (match (pre-order-traversal-recursive child index)
                         ((next-index traversed-child)
                          (list next-index (cons traversed-child result))))))
                    (list (1+ index) (list))
                    children)
         ((next-index traversed-children)
          (list next-index
                (cons index (reverse traversed-children))))))
      (leaf (list (1+ index) index))))

  (match (pre-order-traversal-recursive tree 0)
    ((_ result) result)))

(define (tree-label tree n)
  (list-ref (nodeset tree) n))

(define (tree-ref tree n)
  "Return the subtree in pre-order traversed TREE with node N as
root. N is the index of the node, starting from 0, for a pre-order
traversal of TREE."
  (define (tree-ref-recursive tree n)
    (match tree
      ((parent children ...)
       (if (= parent n)
           tree
           (any (cut tree-ref-recursive <> n)
                children)))
      (leaf (if (= leaf n) leaf #f))))

  (or (tree-ref-recursive tree n)
      (error "Subtree not found" tree n)))

(define (tree-root tree)
  (match tree
    ((root children ...) root)
    (root root)))

(define (tree-parent-ref tree n)
  "Return the parent of subtree in pre-order traversed TREE with node
N as root. N is the index of the node, starting from 0, for a
pre-order traversal of TREE."
  (define (tree-parent-ref-recursive tree n)
    (match tree
      ((parent children ...)
       (if (any (lambda (child)
                  (= (tree-root child) n))
                children)
           parent
           (any (cut tree-parent-ref-recursive <> n)
                children)))
      (leaf #f)))

  (and (not (zero? n))
       (or (tree-parent-ref-recursive tree n)
           (error "Parent not found" tree n))))

(define (children tree)
  (match tree
    ((_ children ...) children)
    (leaf (list))))

(define (child-nodeset tree)
  (append-map nodeset (children tree)))

(define (edge-exists? adjacency m n)
  "Return #t if edge from node M to node N exists in ADJACENCY
matrix."
  (match (array-ref adjacency m n)
    ((lower-bound . upper-bound)
     (and (finite? lower-bound)
          (finite? upper-bound)))))

(define (cost-forced-move-left m* n adjacency traversed1 traversed2)
  "Return the cost of a forced move from M* (in TRAVERSED1) to N (in
TRAVERSED2) for ADJACENCY matrix given TRAVERSED1 and TRAVERSED2 are
the initial and final traverseds."
  (if (any (cut edge-exists? adjacency m* <>)
           (child-nodeset (tree-ref traversed2 n)))
      0
      cost-move))

(define (cost-forced-move-right m n* adjacency traversed1 traversed2)
  "Return the cost of a forced move from N* (in TRAVERSED2) to M (in
TRAVERSED1) for ADJACENCY matrix given TRAVERSED1 and TRAVERSED2 are
the pre-order traversed initial and final trees."
  (if (any (cut edge-exists? adjacency <> n*)
           (child-nodeset (tree-ref traversed1 m)))
      0
      cost-move))

;; TODO: This lower bound cost is a little different from that in the
;; MH-DIFF paper. Reconcile them properly.
(define (cost-lower-bound m n adjacency tree1 tree2 traversed1 traversed2)
  "Return the lower bound on the cost of the edge from node M to node
N for ADJACENCY matrix given TREE1 and TREE2 are the initial and final
trees respectively. TRAVERSED1 and TRAVERSED2 are the pre-order
traversed initial and final trees."
  (cond
   ;; These special cases for + and - are missing in the MH-Diff paper.
   ((and (eq? m '+) (eq? n '-)) 0)
   ((eq? m '+) cost-insert)
   ((eq? n '-) cost-delete)
   (else
    (+ (cost-update m n tree1 tree2)
       (/ (+ (list-transduce (tmap (cut cost-forced-move-left <> n adjacency traversed1 traversed2))
                             +
                             (child-nodeset (tree-ref traversed1 m)))
             (list-transduce (tmap (cut cost-forced-move-right m <> adjacency traversed1 traversed2))
                             +
                             (child-nodeset (tree-ref traversed2 n))))
          2))
    ;; This is the original lower bound cost according to the paper.
    ;; (+ (cost-update m n tree1 tree2)
    ;;    (/ (min (list-transduce (tmap (cut cost-forced-move-left <> n adjacency traversed1 traversed2))
    ;;                            +
    ;;                            (child-nodeset (tree-ref traversed1 m)))
    ;;            (list-transduce (tmap (cut cost-forced-move-right m <> adjacency traversed1 traversed2))
    ;;                            +
    ;;                            (child-nodeset (tree-ref traversed2 n))))
    ;;       2))
    )))

;; TODO: The definition of the cost-conditional-move-* in the MH-DIFF
;; paper is identical to that of cost-forced-move-*. Figure out why
;; that is so.
(define cost-conditional-move-left cost-forced-move-left)
(define cost-conditional-move-right cost-forced-move-right)

(define (cost-w m n tree1 tree2)
  (cond
   ((and (eq? m '+) (eq? n '-)) 0)
   ((and (eq? m '+) (not (eq? n '-))) cost-insert)
   ((and (not (eq? m '+)) (eq? n '-)) cost-delete)
   (else (cost-update m n tree1 tree2))))

(define (edge-count sub-adjacency)
  "Return the number of edges in SUB-ADJACENCY matrix."
  (matrix-fold (lambda (i j element result)
                 (if (edge-exists? sub-adjacency i j)
                     (1+ result)
                     result))
               0
               sub-adjacency))

(define (left-incident-edge-count adjacency m)
  "Return the number of edges in ADJACENCY that are left-incident on
node M."
  (edge-count (matrix-row adjacency m)))

(define (right-incident-edge-count adjacency n)
  "Return the number of edges in ADJACENCY that are right-incident on
node N."
  (edge-count (matrix-col adjacency n)))

(define (cost-upper-bound m n adjacency tree1 tree2 traversed1 traversed2)
  "Return the upper bound on the cost of the edge from node M to node
N for ADJACENCY matrix given TREE1 and TREE2 are the initial and final
trees respectively. TRAVERSED1 and TRAVERSED2 are the pre-order
traversed initial and final trees."
  (cond
   ;; TODO: Where are these special cases for + and - in the MH-diff
   ;; paper?
   ((and (eq? m '+) (eq? n '-)) 0)
   ((eq? m '+) cost-insert)
   ((eq? n '-) cost-delete)
   (else
    (+ (cost-w m n tree1 tree2)
       (/ (+ (list-transduce (tmap (lambda (m*)
                                     (+ (* cost-copy
                                           (1- (left-incident-edge-count adjacency m*)))
                                        (cost-conditional-move-left m* n adjacency traversed1 traversed2))))
                             +
                             (child-nodeset (tree-ref traversed1 m)))
             (list-transduce (tmap (lambda (n*)
                                     (+ (* cost-glue
                                           (1- (right-incident-edge-count adjacency n*)))
                                        (cost-conditional-move-right m n* adjacency traversed1 traversed2))))
                             +
                             (child-nodeset (tree-ref traversed2 n))))
          2)))))

(define (compute-costs! adjacency tree1 tree2)
  "Fill ADJACENCY matrix with lower and upper bound costs. Return the
filled matrix. TREE1 and TREE2 are the initial and final
trees. TRAVERSED1 and TRAVERSED2 are the pre-order traversed initial
and final trees."
  (let ((traversed1 (pre-order-traversal tree1))
        (traversed2 (pre-order-traversal tree2)))
    (array-index-map! adjacency
                      (lambda (i j)
                        (match (array-ref adjacency i j)
                          ((+inf.0 . +inf.0) (cons +inf.0 +inf.0))
                          (_
                           ;; The last nodes are the special + and -
                           ;; nodes.
                           (let ((m (if (< i (length (nodeset tree1))) i '+))
                                 (n (if (< j (length (nodeset tree2))) j '-)))
                             (cons (cost-lower-bound m n adjacency tree1 tree2 traversed1 traversed2)
                                   (cost-upper-bound m n adjacency tree1 tree2 traversed1 traversed2))))))))
  adjacency)

(define (induced-graph nodeset1 nodeset2 tree1 tree2)
  "Return the adjacency matrix of the induced graph corresponding to
initial tree TREE1 and final tree TREE2."
  ;; The cost of each edge depends only on the existence of other
  ;; edges, not on the cost of other edges. So, we first fill the
  ;; adjacency matrix with zeros to indicate that every edge exists.
  (compute-costs! (make-array (cons 0 0)
                              (1+ (length nodeset1))
                              (1+ (length nodeset2)))
                  tree1 tree2))

(define (prune? m n adjacency)
  "Return #t if edge from M to N in ADJACENCY should be pruned, else
return #f."
  (let* ((lower-bound-cost
          (match (array-ref adjacency m n)
            ((lower-bound-cost . _) lower-bound-cost)))
         (upper-bound-cost (match-lambda ((_ . upper-bound) upper-bound)))
         (minimum-upper-bound-cost
          (lambda (cost1 cost2)
            (if (< (upper-bound-cost cost1)
                   (upper-bound-cost cost2))
                cost1 cost2))))
    (or
     ;; Pruning rule 1
     (>= lower-bound-cost
         (+ (upper-bound-cost (matrix-min (matrix-row adjacency m)
                                          minimum-upper-bound-cost))
            (upper-bound-cost (matrix-min (matrix-col adjacency n)
                                          minimum-upper-bound-cost))
            (* 2 (max cost-move cost-copy cost-glue))))
     ;; Pruning rule 2
     (>= lower-bound-cost
         (+ cost-delete cost-insert)))))

(define (prune adjacency tree1 tree2)
  "Prune edge cover represented by ADJACENCY matrix. TREE1 and TREE2
are the initial and final trees."
  (let ((adjacency-original (array-copy adjacency)))
    (array-index-map! adjacency (lambda (m n)
                                  (if (prune? m n adjacency)
                                      (cons +inf.0 +inf.0)
                                      (array-ref adjacency m n))))
    (compute-costs! adjacency tree1 tree2)
    (if (array-equal? adjacency-original adjacency)
        adjacency
        (prune adjacency tree1 tree2))))

(define (lower-bound-adjacency adjacency)
  (let ((result (apply make-array 0 (array-dimensions adjacency))))
    (array-map! result
                (match-lambda
                  ((lower-bound . _) lower-bound))
                adjacency)
    result))

(define (edit-script edge-cover nodeset1 nodeset2 tree1 tree2 traversed1 traversed2)
  (filter-map (match-lambda
                (('+ . '-) #f)
                (('+ . n) (list 'insert n))
                ((m . '-) (list 'delete m))
                ((m . n) (and (not (equal? (tree-label tree1 m)
                                           (tree-label tree2 n)))
                              (list 'update m n))))
              edge-cover))

(define (annotate-old-tree tree traversed edit-script)
  (define (annotate-node node)
    (or (any (match-lambda
               (('delete m)
                (and (= m node)
                     (make <sdiff-node>
                       #:tag 'delete
                       #:label (tree-label tree node))))
               (('update m n)
                (and (= m node)
                     (make <sdiff-node>
                       #:tag 'update-old
                       #:label (tree-label tree node))))
               (_ #f))
             edit-script)
        (tree-label tree node)))
  
  (match traversed
    ((parent children ...)
     (cons (annotate-node parent)
           (map (cut annotate-old-tree tree <> edit-script) children)))
    (leaf (annotate-node leaf))))

(define (annotate-new-tree tree traversed edit-script)
  (define (annotate-node node)
    (or (any (match-lambda
               (('insert n)
                (and (= n node)
                     (make <sdiff-node>
                       #:tag 'insert
                       #:label (tree-label tree node))))
               (('update m n)
                (and (= n node)
                     (make <sdiff-node>
                       #:tag 'update-new
                       #:label (tree-label tree node))))
               (_ #f))
             edit-script)
        (tree-label tree node)))
  
  (match traversed
    ((parent children ...)
     (cons (annotate-node parent)
           (map (cut annotate-new-tree tree <> edit-script) children)))
    (leaf (annotate-node leaf))))

(define (sdiff old-tree new-tree)
  (let* ((old-tree (make-proper-tree old-tree))
         (new-tree (make-proper-tree new-tree))
         (traversed1 (pre-order-traversal old-tree))
         (traversed2 (pre-order-traversal new-tree))
         (nodeset1 (nodeset traversed1))
         (nodeset2 (nodeset traversed2))
         (edge-cover
          (map (match-lambda
                 ((i . j)
                  (cons (if (= i (length nodeset1)) '+ i)
                        (if (= j (length nodeset2)) '- j))))
               (minimum-cost-edge-cover
                (lower-bound-adjacency
                 (prune (induced-graph nodeset1 nodeset2 old-tree new-tree)
                        old-tree new-tree)))))
         (edit-script (edit-script edge-cover nodeset1 nodeset2 old-tree new-tree
                                   traversed1 traversed2)))
    (pretty-print (annotate-old-tree old-tree traversed1 edit-script))
    (display "---")
    (newline)
    (pretty-print (annotate-new-tree new-tree traversed2 edit-script))))
